from __future__ import unicode_literals

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.timezone import now

from djangoql.queryset import DjangoQLQuerySet


class Book(models.Model):
    GENRES = {
        1: 'Drama',
        2: 'Comics',
        3: 'Other',
    }

    name = models.CharField(max_length=10)  # lol, we're minimalists
    author = models.ForeignKey('auth.User', on_delete=models.CASCADE)
    genre = models.PositiveIntegerField(
        null=True,
        blank=True,
        choices=GENRES.items(),
    )
    written = models.DateTimeField(default=now)
    is_published = models.BooleanField(default=False)
    rating = models.FloatField(null=True)
    price = models.DecimalField(max_digits=7, decimal_places=2, null=True)
    content_type = models.ForeignKey(
        ContentType,
        null=True,
        on_delete=models.CASCADE,
        editable=False,
    )
    object_id = models.PositiveIntegerField(null=True, editable=False)
    content_object = GenericForeignKey('content_type', 'object_id')

    similar_books = models.ManyToManyField('Book', blank=True, related_name='+')

    objects = DjangoQLQuerySet.as_manager()

    def __str__(self):
        return self.name



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-06-03 08:12
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='book',
            name='genre',
            field=models.PositiveIntegerField(choices=[(1, 'Drama'), (2, 'Comics'), (3, 'Other')], null=True, blank=True),
        ),
    ]



import json

from django.core.paginator import EmptyPage, Paginator
from django.http import HttpResponse
from django.views.generic.base import View


class SuggestionsAPIView(View):
    http_method_names = ['get']
    schema = None
    items_per_page = 100

    def get(self, request, *args, **kwargs):
        search = request.GET.get('search', '')

        try:
            field_name = request.GET.get('field', '')
            field = self.get_field(field_name)
            page_number = int(request.GET.get('page', 1))
            if page_number < 1:
                raise ValueError('page must be an integer starting from 1')
            suggestions = self.get_suggestions(field=field, search=search)
        except ValueError as e:
            error = str(e) or e.__class__.__name__
            return HttpResponse(
                content=json.dumps({'error': error}, indent=2),
                content_type='application/json; charset=utf-8',
                status=400,
            )

        paginator = Paginator(suggestions, self.items_per_page)
        try:
            page = paginator.page(page_number)
        except EmptyPage:
            items = []
            has_next = False
        else:
            items = list(page.object_list)
            has_next = page.has_next()

        response = {
            'items': items,
            'page': page_number,
            'has_next': has_next,
        }
        return HttpResponse(
            content=json.dumps(response, indent=2),
            content_type='application/json; charset=utf-8',
        )

    def get_field(self, field_name):
        if not self.schema:
            raise ValueError('DjangoQL schema is undefined')
        if not field_name:
            raise ValueError('"field" parameter is required')
        parts = field_name.split('.')
        field_name = parts.pop()
        if parts:
            model_name = parts[-1]
            app_label = '.'.join(parts[:-1])
            if not app_label:
                app_label = self.schema.current_model._meta.app_label
            model_label = '.'.join([app_label, model_name])
        else:
            model_label = self.schema.model_label(self.schema.current_model)
        schema_model = self.schema.models.get(model_label)
        if not schema_model:
            raise ValueError('Unknown model: %s' % model_label)
        field_instance = schema_model.get(field_name)
        if not field_instance:
            raise ValueError('Unknown field: %s' % field_name)
        return field_instance

    def get_suggestions(self, field, search):
        if not field.suggest_options:
            raise ValueError("%s.%s doesn't support suggestions" % (
                field.model._meta.object_name,
                field.name,
            ))
        return field.get_options(search)



import inspect
import warnings
from collections import OrderedDict, deque
from datetime import datetime
from decimal import Decimal

from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import ManyToManyRel, ManyToOneRel
from django.db.models.fields.related import ForeignObjectRel
from django.utils.timezone import get_current_timezone

from .ast import Comparison, Const, List, Logical, Name, Node
from .compat import text_type
from .exceptions import DjangoQLSchemaError


class DjangoQLField(object):
    """
    Abstract searchable field
    """
    model = None
    name = None
    nullable = False
    suggest_options = False
    type = 'unknown'
    value_types = []
    value_types_description = ''

    def __init__(self, model=None, name=None, nullable=None,
                 suggest_options=None):
        if model is not None:
            self.model = model
        if name is not None:
            self.name = name
        if nullable is not None:
            self.nullable = nullable
        if suggest_options is not None:
            self.suggest_options = suggest_options

    def _field_choices(self):
        if self.model:
            try:
                return self.model._meta.get_field(self.name).choices
            except (AttributeError, FieldDoesNotExist):
                pass
        return []

    @property
    def async_options(self):
        return not self._field_choices()

    def get_options(self, search):
        """
        Override this method to provide custom suggestion options
        """
        result = []
        choices = self._field_choices()
        if choices:
            search = search.lower()
            for c in choices:
                choice = text_type(c[1])
                if search in choice.lower():
                    result.append(choice)
        return result

    def get_lookup_name(self):
        """
        Override this method to provide custom lookup name
        """
        return self.name

    def get_lookup_value(self, value):
        """
        Override this method to convert displayed values to lookup values
        """
        choices = self._field_choices()
        if choices:
            if isinstance(value, list):
                return [c[0] for c in choices if c[0] in value or c[1] in value]
            else:
                for c in choices:
                    if value in c:
                        return c[0]
        return value

    def get_operator(self, operator):
        """
        Get a comparison suffix to be used in Django ORM & inversion flag for it

        :param operator: string, DjangoQL comparison operator
        :return: (suffix, invert) - a tuple with 2 values:
            suffix - suffix to be used in ORM query, for example '__gt' for '>'
            invert - boolean, True if this comparison needs to be inverted
        """
        op = {
            '=': '',
            '>': '__gt',
            '>=': '__gte',
            '<': '__lt',
            '<=': '__lte',
            '~': '__icontains',
            'in': '__in',
            'startswith': '__istartswith',
            'endswith': '__iendswith',
        }.get(operator)
        if op is not None:
            return op, False
        op = {
            '!=': '',
            '!~': '__icontains',
            'not in': '__in',
            'not startswith': '__istartswith',
            'not endswith': '__iendswith',
        }[operator]
        return op, True

    def get_lookup(self, path, operator, value):
        """
        Performs a lookup for this field with given path, operator and value.

        Override this if you'd like to implement a fully custom lookup. It
        should support all comparison operators compatible with the field type.

        :param path: a list of names preceding current lookup. For example,
            if expression looks like 'author.groups.name = "Foo"' path would
            be ['author', 'groups']. 'name' is not included, because it's the
            current field instance itself.
        :param operator: a string with comparison operator. It could be one of
            the following: '=', '!=', '>', '>=', '<', '<=', '~', '!~', 'in',
            'not in'. Depending on the field type, some operators may be
            excluded. '~' and '!~' can be applied to StrField only and aren't
            allowed for any other fields. BoolField can't be used with less or
            greater operators, '>', '>=', '<' and '<=' are excluded for it.
        :param value: value passed for comparison
        :return: Q-object
        """
        search = '__'.join(path + [self.get_lookup_name()])
        op, invert = self.get_operator(operator)
        q = models.Q(**{'%s%s' % (search, op): self.get_lookup_value(value)})
        return ~q if invert else q

    def validate(self, value):
        if not self.nullable and value is None:
            raise DjangoQLSchemaError(
                'Field %s is not nullable, '
                "can't compare it to None" % self.name,
            )
        if value is not None and type(value) not in self.value_types:
            if self.nullable:
                msg = (
                    'Field "{field}" has "nullable {field_type}" type. '
                    'It can be compared to {possible_values} or None, '
                    'but not to {value}'
                )
            else:
                msg = (
                    'Field "{field}" has "{field_type}" type. It can '
                    'be compared to {possible_values}, '
                    'but not to {value}'
                )
            raise DjangoQLSchemaError(msg.format(
                field=self.name,
                field_type=self.type,
                possible_values=self.value_types_description,
                value=repr(value),
            ))


class IntField(DjangoQLField):
    type = 'int'
    value_types = [int]
    value_types_description = 'integer numbers'

    def validate(self, value):
        """
        Support enum-like choices defined on an integer field
        """
        return super(IntField, self).validate(self.get_lookup_value(value))


class FloatField(DjangoQLField):
    type = 'float'
    value_types = [int, float, Decimal]
    value_types_description = 'floating point numbers'


class StrField(DjangoQLField):
    type = 'str'
    value_types = [text_type]
    value_types_description = 'strings'

    def get_options(self, search):
        choice_options = super(StrField, self).get_options(search)
        if choice_options:
            return choice_options
        lookup = {}
        if search:
            lookup['%s__icontains' % self.name] = search
        return self.model.objects\
            .filter(**lookup)\
            .order_by(self.name)\
            .values_list(self.name, flat=True)\
            .distinct()


class BoolField(DjangoQLField):
    type = 'bool'
    value_types = [bool]
    value_types_description = 'True or False'


class DateField(DjangoQLField):
    type = 'date'
    value_types = [text_type]
    value_types_description = 'dates in "YYYY-MM-DD" format'

    def validate(self, value):
        super(DateField, self).validate(value)
        try:
            self.get_lookup_value(value)
        except ValueError:
            raise DjangoQLSchemaError(
                'Field "%s" can be compared to dates in '
                '"YYYY-MM-DD" format, but not to %s' % (
                    self.name,
                    repr(value),
                ),
            )

    def get_lookup_value(self, value):
        return (
            [self.parse_date(v) for v in value]
            if isinstance(value, list)
            else self.parse_date(value)
        )

    @staticmethod
    def parse_date(value):
        if not value:
            return None
        return datetime.strptime(value, '%Y-%m-%d').date()


class DateTimeField(DjangoQLField):
    type = 'datetime'
    value_types = [text_type]
    value_types_description = 'timestamps in "YYYY-MM-DD HH:MM" format'

    def validate(self, value):
        super(DateTimeField, self).validate(value)
        try:
            self.get_lookup_value(value)
        except ValueError:
            raise DjangoQLSchemaError(
                'Field "%s" can be compared to timestamps in '
                '"YYYY-MM-DD HH:MM" format, but not to %s' % (
                    self.name,
                    repr(value),
                ),
            )

    def get_lookup_value(self, value):
        return (
            [self.parse_datetime(v) for v in value]
            if isinstance(value, list)
            else self.parse_datetime(value)
        )

    @staticmethod
    def parse_datetime(value):
        if not value:
            return None
        mask = '%Y-%m-%d'
        if len(value) > 10:
            mask += ' %H:%M'
        if len(value) > 16:
            mask += ':%S'
        dt = datetime.strptime(value, mask)
        if settings.USE_TZ:
            dt = dt.replace(tzinfo=get_current_timezone())
        return dt

    def get_lookup(self, path, operator, value):
        search = '__'.join(path + [self.get_lookup_name()])
        op, invert = self.get_operator(operator)

        # Add LIKE operator support for datetime fields. For LIKE comparisons
        # we don't want to convert source value to datetime instance, because
        # it would effectively kill the idea. What we want is expressions like
        #       'created ~ "2017-01-30'
        # to be translated to
        #       'created LIKE %2017-01-30%',
        # but it would work only if we pass a string as a parameter. If we pass
        # a datetime instance, it would add time part in a form of 00:00:00,
        # and resulting comparison would look like
        #       'created LIKE %2017-01-30 00:00:00%'
        # which is not what we want for this case.
        val = value if operator in ('~', '!~') else self.get_lookup_value(value)

        q = models.Q(**{'%s%s' % (search, op): val})
        return ~q if invert else q


class RelationField(DjangoQLField):
    type = 'relation'

    def __init__(self, model, name, related_model, nullable=False,
                 suggest_options=False):
        super(RelationField, self).__init__(
            model=model,
            name=name,
            nullable=nullable,
            suggest_options=suggest_options,
        )
        self.related_model = related_model

    @property
    def relation(self):
        return DjangoQLSchema.model_label(self.related_model)


class DjangoQLSchema(object):
    include = ()  # models to include into introspection
    exclude = ()  # models to exclude from introspection
    suggest_options = None

    def __init__(self, model):
        if not inspect.isclass(model) or not issubclass(model, models.Model):
            raise DjangoQLSchemaError(
                'Schema must be initialized with a subclass of Django model',
            )
        if self.include and self.exclude:
            raise DjangoQLSchemaError(
                'Either include or exclude can be specified, but not both',
            )
        if self.excluded(model):
            raise DjangoQLSchemaError(
                "%s can't be used with %s because it's excluded from it" % (
                    model,
                    self.__class__,
                ),
            )
        self.current_model = model
        self._models = None
        if self.suggest_options is None:
            self.suggest_options = {}

    def excluded(self, model):
        return model in self.exclude or (
            self.include and model not in self.include
        )

    @property
    def models(self):
        if not self._models:
            self._models = self.introspect(
                model=self.current_model,
                exclude=tuple(self.model_label(m) for m in self.exclude),
            )
        return self._models

    @classmethod
    def model_label(self, model):
        return text_type(model._meta)

    def introspect(self, model, exclude=()):
        """
        Start with given model and recursively walk through its relationships.

        Returns a dict with all model labels and their fields found.
        """
        result = {}
        open_set = deque([model])
        closed_set = set(exclude)

        while open_set:
            model = open_set.popleft()
            model_label = self.model_label(model)

            if model_label in closed_set:
                continue

            model_fields = OrderedDict()
            for field in self.get_fields(model):
                if not isinstance(field, DjangoQLField):
                    field = self.get_field_instance(model, field)
                if not field:
                    continue
                if isinstance(field, RelationField):
                    open_set.append(field.related_model)
                model_fields[field.name] = field

            result[model_label] = model_fields
            closed_set.add(model_label)

        return result

    def get_fields(self, model):
        """
        By default, returns all field names of a given model.

        Override this method to limit field options. You can either return a
        plain list of field names from it, like ['id', 'name'], or call
        .super() and exclude unwanted fields from its result.
        """
        return sorted(
            [f.name for f in model._meta.get_fields() if f.name != 'password'],
        )

    def get_field_instance(self, model, field_name):
        field = model._meta.get_field(field_name)
        field_kwargs = {'model': model, 'name': field.name}
        if field.is_relation:
            if not field.related_model:
                # GenericForeignKey
                return
            if self.excluded(field.related_model):
                return
            field_cls = RelationField
            field_kwargs['related_model'] = field.related_model
        else:
            field_cls = self.get_field_cls(field)
        if isinstance(field, (ManyToOneRel, ManyToManyRel, ForeignObjectRel)):
            # Django 1.8 doesn't have .null attribute for these fields
            field_kwargs['nullable'] = True
        else:
            field_kwargs['nullable'] = field.null
        field_kwargs['suggest_options'] = (
            field.name in self.suggest_options.get(model, [])
        )
        return field_cls(**field_kwargs)

    def get_field_cls(self, field):
        str_fields = (
            models.CharField,
            models.TextField,
            models.UUIDField,
            models.BinaryField,
            models.GenericIPAddressField,
        )
        if isinstance(field, str_fields):
            return StrField
        elif isinstance(field, (models.AutoField, models.IntegerField)):
            return IntField
        elif isinstance(field, (models.BooleanField, models.NullBooleanField)):
            return BoolField
        elif isinstance(field, (models.DecimalField, models.FloatField)):
            return FloatField
        elif isinstance(field, models.DateTimeField):
            return DateTimeField
        elif isinstance(field, models.DateField):
            return DateField
        return DjangoQLField

    def as_dict(self):
        from .serializers import DjangoQLSchemaSerializer
        warnings.warn(
            'DjangoQLSchema.as_dict() is deprecated and will be removed in '
            'future releases. Please use DjangoQLSchemaSerializer instead.',
        )
        return DjangoQLSchemaSerializer().serialize(self)

    def resolve_name(self, name):
        assert isinstance(name, Name)
        model = self.model_label(self.current_model)
        field = None
        for name_part in name.parts:
            field = self.models[model].get(name_part)
            if not field:
                raise DjangoQLSchemaError(
                    'Unknown field: %s. Possible choices are: %s' % (
                        name_part,
                        ', '.join(sorted(self.models[model].keys())),
                    ),
                )
            if field.type == 'relation':
                model = field.relation
                field = None
        return field

    def validate(self, node):
        """
        Validate DjangoQL AST tree vs. current schema
        """
        assert isinstance(node, Node)
        if isinstance(node.operator, Logical):
            self.validate(node.left)
            self.validate(node.right)
            return
        assert isinstance(node.left, Name)
        assert isinstance(node.operator, Comparison)
        assert isinstance(node.right, (Const, List))

        # Check that field and value types are compatible
        field = self.resolve_name(node.left)
        value = node.right.value
        if field is None:
            if value is not None:
                raise DjangoQLSchemaError(
                    'Related model %s can be compared to None only, but not to '
                    '%s' % (node.left.value, type(value).__name__),
                )
        else:
            values = value if isinstance(node.right, List) else [value]
            for v in values:
                field.validate(v)




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'expressionAND COMMA CONTAINS EQUALS FALSE FLOAT_VALUE GREATER GREATER_EQUAL IN INT_VALUE LESS LESS_EQUAL NAME NONE NOT NOT_CONTAINS NOT_EQUALS OR PAREN_L PAREN_R STRING_VALUE TRUE\n        expression : PAREN_L expression PAREN_R\n        \n        expression : expression logical expression\n        \n        expression : name comparison_number number\n                   | name comparison_string string\n                   | name comparison_equality boolean_value\n                   | name comparison_equality none\n                   | name comparison_in_list const_list_value\n        \n        name : NAME\n        \n        logical : AND\n                | OR\n        \n        comparison_number : comparison_equality\n                          | comparison_greater_less\n        \n        comparison_string : comparison_equality\n                          | comparison_greater_less\n                          | comparison_contains\n        \n        comparison_equality : EQUALS\n                            | NOT_EQUALS\n        \n        comparison_greater_less : GREATER\n                                | GREATER_EQUAL\n                                | LESS\n                                | LESS_EQUAL\n        \n        comparison_contains : CONTAINS\n                            | NOT_CONTAINS\n        \n        comparison_in_list : IN\n                           | NOT IN\n        \n        const_value : number\n                    | string\n                    | none\n                    | boolean_value\n        \n        number : INT_VALUE\n        \n        number : FLOAT_VALUE\n        \n        string : STRING_VALUE\n        \n        none : NONE\n        \n        boolean_value : true\n                      | false\n        \n        true : TRUE\n        \n        false : FALSE\n        \n        const_list_value : PAREN_L const_value_list PAREN_R\n        \n        const_value_list : const_value_list COMMA const_value\n        \n        const_value_list : const_value\n        '
    
_lr_action_items = {'PAREN_L':([0,2,5,6,7,12,17,41,],[2,2,2,-9,-10,40,-24,-25,]),'NAME':([0,2,5,6,7,],[4,4,4,-9,-10,]),'$end':([1,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,48,],[0,-2,-1,-3,-30,-31,-4,-32,-5,-6,-34,-35,-33,-36,-37,-7,-38,]),'AND':([1,8,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,48,],[6,6,6,-1,-3,-30,-31,-4,-32,-5,-6,-34,-35,-33,-36,-37,-7,-38,]),'OR':([1,8,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,48,],[7,7,7,-1,-3,-30,-31,-4,-32,-5,-6,-34,-35,-33,-36,-37,-7,-38,]),'EQUALS':([3,4,],[15,-8,]),'NOT_EQUALS':([3,4,],[16,-8,]),'IN':([3,4,18,],[17,-8,41,]),'NOT':([3,4,],[18,-8,]),'GREATER':([3,4,],[19,-8,]),'GREATER_EQUAL':([3,4,],[20,-8,]),'LESS':([3,4,],[21,-8,]),'LESS_EQUAL':([3,4,],[22,-8,]),'CONTAINS':([3,4,],[23,-8,]),'NOT_CONTAINS':([3,4,],[24,-8,]),'PAREN_R':([8,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,42,43,44,45,46,47,48,50,],[26,-2,-1,-3,-30,-31,-4,-32,-5,-6,-34,-35,-33,-36,-37,-7,48,-40,-26,-27,-28,-29,-38,-39,]),'INT_VALUE':([9,11,13,15,16,19,20,21,22,40,49,],[28,-11,-12,-16,-17,-18,-19,-20,-21,28,28,]),'FLOAT_VALUE':([9,11,13,15,16,19,20,21,22,40,49,],[29,-11,-12,-16,-17,-18,-19,-20,-21,29,29,]),'STRING_VALUE':([10,11,13,14,15,16,19,20,21,22,23,24,40,49,],[31,-13,-14,-15,-16,-17,-18,-19,-20,-21,-22,-23,31,31,]),'NONE':([11,15,16,40,49,],[36,-16,-17,36,36,]),'TRUE':([11,15,16,40,49,],[37,-16,-17,37,37,]),'FALSE':([11,15,16,40,49,],[38,-16,-17,38,38,]),'COMMA':([28,29,31,34,35,36,37,38,42,43,44,45,46,47,50,],[-30,-31,-32,-34,-35,-33,-36,-37,49,-40,-26,-27,-28,-29,-39,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expression':([0,2,5,],[1,8,25,]),'name':([0,2,5,],[3,3,3,]),'logical':([1,8,25,],[5,5,5,]),'comparison_number':([3,],[9,]),'comparison_string':([3,],[10,]),'comparison_equality':([3,],[11,]),'comparison_in_list':([3,],[12,]),'comparison_greater_less':([3,],[13,]),'comparison_contains':([3,],[14,]),'number':([9,40,49,],[27,44,44,]),'string':([10,40,49,],[30,45,45,]),'boolean_value':([11,40,49,],[32,47,47,]),'none':([11,40,49,],[33,46,46,]),'true':([11,40,49,],[34,34,34,]),'false':([11,40,49,],[35,35,35,]),'const_list_value':([12,],[39,]),'const_value_list':([40,],[42,]),'const_value':([40,49,],[43,50,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expression","S'",1,None,None,None),
  ('expression -> PAREN_L expression PAREN_R','expression',3,'p_expression_parens','parser.py',49),
  ('expression -> expression logical expression','expression',3,'p_expression_logical','parser.py',55),
  ('expression -> name comparison_number number','expression',3,'p_expression_comparison','parser.py',61),
  ('expression -> name comparison_string string','expression',3,'p_expression_comparison','parser.py',62),
  ('expression -> name comparison_equality boolean_value','expression',3,'p_expression_comparison','parser.py',63),
  ('expression -> name comparison_equality none','expression',3,'p_expression_comparison','parser.py',64),
  ('expression -> name comparison_in_list const_list_value','expression',3,'p_expression_comparison','parser.py',65),
  ('name -> NAME','name',1,'p_name','parser.py',71),
  ('logical -> AND','logical',1,'p_logical','parser.py',77),
  ('logical -> OR','logical',1,'p_logical','parser.py',78),
  ('comparison_number -> comparison_equality','comparison_number',1,'p_comparison_number','parser.py',84),
  ('comparison_number -> comparison_greater_less','comparison_number',1,'p_comparison_number','parser.py',85),
  ('comparison_string -> comparison_equality','comparison_string',1,'p_comparison_string','parser.py',91),
  ('comparison_string -> comparison_greater_less','comparison_string',1,'p_comparison_string','parser.py',92),
  ('comparison_string -> comparison_contains','comparison_string',1,'p_comparison_string','parser.py',93),
  ('comparison_equality -> EQUALS','comparison_equality',1,'p_comparison_equality','parser.py',99),
  ('comparison_equality -> NOT_EQUALS','comparison_equality',1,'p_comparison_equality','parser.py',100),
  ('comparison_greater_less -> GREATER','comparison_greater_less',1,'p_comparison_greater_less','parser.py',106),
  ('comparison_greater_less -> GREATER_EQUAL','comparison_greater_less',1,'p_comparison_greater_less','parser.py',107),
  ('comparison_greater_less -> LESS','comparison_greater_less',1,'p_comparison_greater_less','parser.py',108),
  ('comparison_greater_less -> LESS_EQUAL','comparison_greater_less',1,'p_comparison_greater_less','parser.py',109),
  ('comparison_contains -> CONTAINS','comparison_contains',1,'p_comparison_contains','parser.py',115),
  ('comparison_contains -> NOT_CONTAINS','comparison_contains',1,'p_comparison_contains','parser.py',116),
  ('comparison_in_list -> IN','comparison_in_list',1,'p_comparison_in_list','parser.py',122),
  ('comparison_in_list -> NOT IN','comparison_in_list',2,'p_comparison_in_list','parser.py',123),
  ('const_value -> number','const_value',1,'p_const_value','parser.py',132),
  ('const_value -> string','const_value',1,'p_const_value','parser.py',133),
  ('const_value -> none','const_value',1,'p_const_value','parser.py',134),
  ('const_value -> boolean_value','const_value',1,'p_const_value','parser.py',135),
  ('number -> INT_VALUE','number',1,'p_number_int','parser.py',141),
  ('number -> FLOAT_VALUE','number',1,'p_number_float','parser.py',147),
  ('string -> STRING_VALUE','string',1,'p_string','parser.py',153),
  ('none -> NONE','none',1,'p_none','parser.py',159),
  ('boolean_value -> true','boolean_value',1,'p_boolean_value','parser.py',165),
  ('boolean_value -> false','boolean_value',1,'p_boolean_value','parser.py',166),
  ('true -> TRUE','true',1,'p_true','parser.py',172),
  ('false -> FALSE','false',1,'p_false','parser.py',178),
  ('const_list_value -> PAREN_L const_value_list PAREN_R','const_list_value',3,'p_const_list_value','parser.py',184),
  ('const_value_list -> const_value_list COMMA const_value','const_value_list',3,'p_const_value_list','parser.py',190),
  ('const_value_list -> const_value','const_value_list',1,'p_const_value_list_single','parser.py',196),
]



from __future__ import unicode_literals

import re
from decimal import Decimal

import ply.yacc as yacc

from .ast import Comparison, Const, Expression, List, Logical, Name
from .compat import binary_type, text_type
from .exceptions import DjangoQLParserError
from .lexer import DjangoQLLexer


unescape_pattern = re.compile(
    '(' + DjangoQLLexer.re_escaped_char + '|' +
    DjangoQLLexer.re_escaped_unicode + ')',
)


def unescape_repl(m):
    contents = m.group(1)
    if len(contents) == 2:
        return contents[1]
    else:
        return contents.encode('utf8').decode('unicode_escape')


def unescape(value):
    if isinstance(value, binary_type):
        value = value.decode('utf8')
    return re.sub(unescape_pattern, unescape_repl, value)


class DjangoQLParser(object):
    def __init__(self, debug=False, **kwargs):
        self.default_lexer = DjangoQLLexer()
        self.tokens = self.default_lexer.tokens
        kwargs['debug'] = debug
        if 'write_tables' not in kwargs:
            kwargs['write_tables'] = False
        self.yacc = yacc.yacc(module=self, **kwargs)

    def parse(self, input=None, lexer=None, **kwargs):  # noqa: A002
        lexer = lexer or self.default_lexer
        return self.yacc.parse(input=input, lexer=lexer, **kwargs)

    start = 'expression'

    def p_expression_parens(self, p):
        """
        expression : PAREN_L expression PAREN_R
        """
        p[0] = p[2]

    def p_expression_logical(self, p):
        """
        expression : expression logical expression
        """
        p[0] = Expression(left=p[1], operator=p[2], right=p[3])

    def p_expression_comparison(self, p):
        """
        expression : name comparison_number number
                   | name comparison_string string
                   | name comparison_equality boolean_value
                   | name comparison_equality none
                   | name comparison_in_list const_list_value
        """
        p[0] = Expression(left=p[1], operator=p[2], right=p[3])

    def p_name(self, p):
        """
        name : NAME
        """
        p[0] = Name(parts=p[1].split('.'))

    def p_logical(self, p):
        """
        logical : AND
                | OR
        """
        p[0] = Logical(operator=p[1])

    def p_comparison_number(self, p):
        """
        comparison_number : comparison_equality
                          | comparison_greater_less
        """
        p[0] = p[1]

    def p_comparison_string(self, p):
        """
        comparison_string : comparison_equality
                          | comparison_greater_less
                          | comparison_string_specific
        """
        p[0] = p[1]

    def p_comparison_equality(self, p):
        """
        comparison_equality : EQUALS
                            | NOT_EQUALS
        """
        p[0] = Comparison(operator=p[1])

    def p_comparison_greater_less(self, p):
        """
        comparison_greater_less : GREATER
                                | GREATER_EQUAL
                                | LESS
                                | LESS_EQUAL
        """
        p[0] = Comparison(operator=p[1])

    def p_comparison_string_specific(self, p):
        """
        comparison_string_specific : CONTAINS
                                   | NOT_CONTAINS
                                   | STARTSWITH
                                   | NOT STARTSWITH
                                   | ENDSWITH
                                   | NOT ENDSWITH
        """
        if len(p) == 2:
            p[0] = Comparison(operator=p[1])
        else:
            p[0] = Comparison(operator='%s %s' % (p[1], p[2]))

    def p_comparison_in_list(self, p):
        """
        comparison_in_list : IN
                           | NOT IN
        """
        if len(p) == 2:
            p[0] = Comparison(operator=p[1])
        else:
            p[0] = Comparison(operator='%s %s' % (p[1], p[2]))

    def p_const_value(self, p):
        """
        const_value : number
                    | string
                    | none
                    | boolean_value
        """
        p[0] = p[1]

    def p_number_int(self, p):
        """
        number : INT_VALUE
        """
        p[0] = Const(value=int(p[1]))

    def p_number_float(self, p):
        """
        number : FLOAT_VALUE
        """
        p[0] = Const(value=Decimal(p[1]))

    def p_string(self, p):
        """
        string : STRING_VALUE
        """
        p[0] = Const(value=unescape(p[1]))

    def p_none(self, p):
        """
        none : NONE
        """
        p[0] = Const(value=None)

    def p_boolean_value(self, p):
        """
        boolean_value : true
                      | false
        """
        p[0] = p[1]

    def p_true(self, p):
        """
        true : TRUE
        """
        p[0] = Const(value=True)

    def p_false(self, p):
        """
        false : FALSE
        """
        p[0] = Const(value=False)

    def p_const_list_value(self, p):
        """
        const_list_value : PAREN_L const_value_list PAREN_R
        """
        p[0] = List(items=p[2])

    def p_const_value_list(self, p):
        """
        const_value_list : const_value_list COMMA const_value
        """
        p[0] = p[1] + [p[3]]

    def p_const_value_list_single(self, p):
        """
        const_value_list : const_value
        """
        p[0] = [p[1]]

    def p_error(self, token):
        if token is None:
            self.raise_syntax_error('Unexpected end of input')
        else:
            fragment = text_type(token.value)
            if len(fragment) > 20:
                fragment = fragment[:17] + '...'
            self.raise_syntax_error(
                'Syntax error at %s' % repr(fragment),
                token=token,
            )

    def raise_syntax_error(self, message, token=None):
        if token is None:
            raise DjangoQLParserError(message)
        lexer = token.lexer
        if callable(getattr(lexer, 'find_column', None)):
            column = lexer.find_column(token)
        else:
            column = None
        raise DjangoQLParserError(
            message=message,
            value=token.value,
            line=token.lineno,
            column=column,
        )



# Generated by Django 2.1.2 on 2018-11-20 20:58

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0003_book_similar_books'),
    ]

    operations = [
        migrations.AlterField(
            model_name='book',
            name='content_type',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='book',
            name='object_id',
            field=models.PositiveIntegerField(editable=False, null=True),
        ),
        migrations.AlterField(
            model_name='book',
            name='similar_books',
            field=models.ManyToManyField(blank=True, related_name='_book_similar_books_+', to='core.Book'),
        ),
    ]



from __future__ import unicode_literals

from .compat import text_type


class Node(object):
    def __str__(self):
        children = []
        for k, v in self.__dict__.items():
            if isinstance(v, (list, tuple)):
                v = '[%s]' % ', '.join([text_type(v) for v in v if v])
            children.append('%s=%s' % (k, v))
        return '<%s%s%s>' % (
            self.__class__.__name__,
            ': ' if children else '',
            ', '.join(children),
        )

    __repr__ = __str__

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for k, v in self.__dict__.items():
            if getattr(other, k) != v:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class Expression(Node):
    def __init__(self, left, operator, right):
        self.left = left
        self.operator = operator
        self.right = right


class Name(Node):
    def __init__(self, parts):
        if isinstance(parts, list):
            self.parts = parts
        elif isinstance(parts, tuple):
            self.parts = list(parts)
        else:
            self.parts = [parts]

    @property
    def value(self):
        return '.'.join(self.parts)


class Const(Node):
    def __init__(self, value):
        self.value = value


class List(Node):
    def __init__(self, items):
        self.items = items

    @property
    def value(self):
        return [i.value for i in self.items]


class Operator(Node):
    def __init__(self, operator):
        self.operator = operator


class Logical(Operator):
    pass


class Comparison(Operator):
    pass



from unittest import TestCase

from djangoql.exceptions import DjangoQLLexerError
from djangoql.lexer import DjangoQLLexer


class DjangoQLLexerTest(TestCase):
    lexer = DjangoQLLexer()

    def assert_output(self, lexer, expected):
        actual = list(lexer)
        len_actual = len(actual)
        len_expected = len(expected)
        self.assertEqual(
            len_actual,
            len_expected,
            'Actual output length %s does not match expected length %s\n'
            'Actual: %s\n'
            'Expected: %s' % (len_actual, len_expected, actual, expected),
        )
        for i, token in enumerate(actual):
            self.assertEqual(token.type, expected[i][0])
            self.assertEqual(token.value, expected[i][1])

    def test_punctuator(self):
        self.assert_output(self.lexer.input('('), [('PAREN_L', '(')])
        self.assert_output(self.lexer.input(')'), [('PAREN_R', ')')])
        self.assert_output(self.lexer.input(','), [('COMMA', ',')])
        self.assert_output(self.lexer.input('='), [('EQUALS', '=')])
        self.assert_output(self.lexer.input('!='), [('NOT_EQUALS', '!=')])
        self.assert_output(self.lexer.input('>'), [('GREATER', '>')])
        self.assert_output(self.lexer.input('>='), [('GREATER_EQUAL', '>=')])
        self.assert_output(self.lexer.input('<'), [('LESS', '<')])
        self.assert_output(self.lexer.input('<='), [('LESS_EQUAL', '<=')])
        self.assert_output(self.lexer.input('~'), [('CONTAINS', '~')])
        self.assert_output(self.lexer.input('!~'), [('NOT_CONTAINS', '!~')])

    def test_name(self):
        for name in ('a', 'myVar_42', '__LOL__', '_', '_0'):
            self.assert_output(self.lexer.input(name), [('NAME', name)])

    def test_entity_props(self):
        self.assert_output(self.lexer.input('a.b.c'), [('NAME', 'a.b.c')])
        try:
            list(self.lexer.input('user . group . id'))
            self.fail('Whitespace around dots must raise an exception')
        except DjangoQLLexerError:
            pass
        try:
            list(self.lexer.input('user..id'))
            self.fail('Two dots in a row must raise an exception')
        except DjangoQLLexerError:
            pass

    def test_reserved_words(self):
        reserved = ('True', 'False', 'None', 'or', 'and', 'in', 'not',
                    'startswith', 'endswith')
        for word in reserved:
            self.assert_output(self.lexer.input(word), [(word.upper(), word)])
        # A word made of reserved words should be treated as a name
        for word in ('True_story', 'not_None', 'inspect', 'startswith_in'):
            self.assert_output(self.lexer.input(word), [('NAME', word)])

    def test_int(self):
        for val in ('0', '-0', '42', '-42'):
            self.assert_output(self.lexer.input(val), [('INT_VALUE', val)])

    def test_float(self):
        for val in ('-0.5e+42', '42.0', '2E64', '2.71e-0002'):
            self.assert_output(self.lexer.input(val), [('FLOAT_VALUE', val)])

    def test_string(self):
        for s in ('""', u'""', '"42"', r'"\t\n\u0042 ^"'):
            self.assert_output(
                self.lexer.input(s),
                [('STRING_VALUE', s.strip('"'))],
            )

    def test_illegal_chars(self):
        for s in ('"', '^'):
            try:
                list(self.lexer.input(s))
                self.fail('Illegal char exception not raised for %s' % repr(s))
            except DjangoQLLexerError as e:
                self.assertEqual(1, e.line)
                self.assertEqual(1, e.column)
                self.assertTrue(
                    str(e).startswith('Line 1, col 1: Illegal character'),
                )
                self.assertEqual(s, e.value)

    def test_positional_info(self):
        for i, t in enumerate(self.lexer.input('1\n  3\n    5\n')):
            self.assertEqual(i + 1, t.lineno)
            self.assertEqual(i * 2 + 1, self.lexer.find_column(t))



from __future__ import unicode_literals

import ply.lex as lex
from ply.lex import TOKEN

from .exceptions import DjangoQLLexerError


class DjangoQLLexer(object):
    def __init__(self, **kwargs):
        self._lexer = lex.lex(module=self, **kwargs)
        self.reset()

    def reset(self):
        self.text = ''
        self._lexer.lineno = 1
        return self

    def input(self, s):
        self.reset()
        self.text = s
        self._lexer.input(s)
        return self

    def token(self):
        return self._lexer.token()

    # Iterator interface
    def __iter__(self):
        return self

    def next(self):
        t = self.token()
        if t is None:
            raise StopIteration
        return t

    __next__ = next

    def find_column(self, t):
        """
        Returns token position in current text, starting from 1
        """
        cr = max(
            self.text.rfind(lt, 0, t.lexpos) for lt in self.line_terminators
        )
        if cr == -1:
            return t.lexpos + 1
        return t.lexpos - cr

    whitespace = ' \t\v\f\u00A0'
    line_terminators = '\n\r\u2028\u2029'

    re_line_terminators = r'\n\r\u2028\u2029'

    re_escaped_char = r'\\[\"\\/bfnrt]'
    re_escaped_unicode = r'\\u[0-9A-Fa-f]{4}'
    re_string_char = r'[^\"\\' + re_line_terminators + u']'

    re_int_value = r'(-?0|-?[1-9][0-9]*)'
    re_fraction_part = r'\.[0-9]+'
    re_exponent_part = r'[eE][\+-]?[0-9]+'

    tokens = [
        'COMMA',
        'OR',
        'AND',
        'NOT',
        'IN',
        'TRUE',
        'FALSE',
        'NONE',
        'NAME',
        'STRING_VALUE',
        'FLOAT_VALUE',
        'INT_VALUE',
        'PAREN_L',
        'PAREN_R',
        'EQUALS',
        'NOT_EQUALS',
        'GREATER',
        'GREATER_EQUAL',
        'LESS',
        'LESS_EQUAL',
        'CONTAINS',
        'NOT_CONTAINS',
        'STARTSWITH',
        'ENDSWITH',
    ]

    t_COMMA = ','
    t_PAREN_L = r'\('
    t_PAREN_R = r'\)'
    t_EQUALS = '='
    t_NOT_EQUALS = '!='
    t_GREATER = '>'
    t_GREATER_EQUAL = '>='
    t_LESS = '<'
    t_LESS_EQUAL = '<='
    t_CONTAINS = '~'
    t_NOT_CONTAINS = '!~'

    t_NAME = r'[_A-Za-z][_0-9A-Za-z]*(\.[_A-Za-z][_0-9A-Za-z]*)*'

    t_ignore = whitespace

    @TOKEN(r'\"(' + re_escaped_char +
           '|' + re_escaped_unicode +
           '|' + re_string_char + r')*\"')
    def t_STRING_VALUE(self, t):
        t.value = t.value[1:-1]  # cut leading and trailing quotes ""
        return t

    @TOKEN(re_int_value + re_fraction_part + re_exponent_part + '|' +
           re_int_value + re_fraction_part + '|' +
           re_int_value + re_exponent_part)
    def t_FLOAT_VALUE(self, t):
        return t

    @TOKEN(re_int_value)
    def t_INT_VALUE(self, t):
        return t

    not_followed_by_name = '(?![_0-9A-Za-z])'

    @TOKEN('or' + not_followed_by_name)
    def t_OR(self, t):
        return t

    @TOKEN('and' + not_followed_by_name)
    def t_AND(self, t):
        return t

    @TOKEN('not' + not_followed_by_name)
    def t_NOT(self, t):
        return t

    @TOKEN('in' + not_followed_by_name)
    def t_IN(self, t):
        return t

    @TOKEN('startswith' + not_followed_by_name)
    def t_STARTSWITH(self, t):
        return t

    @TOKEN('endswith' + not_followed_by_name)
    def t_ENDSWITH(self, t):
        return t

    @TOKEN('True' + not_followed_by_name)
    def t_TRUE(self, t):
        return t

    @TOKEN('False' + not_followed_by_name)
    def t_FALSE(self, t):
        return t

    @TOKEN('None' + not_followed_by_name)
    def t_NONE(self, t):
        return t

    def t_error(self, t):
        raise DjangoQLLexerError(
            message='Illegal character %s' % repr(t.value[0]),
            value=t.value,
            line=t.lineno,
            column=self.find_column(t),
        )

    @TOKEN('[' + re_line_terminators + ']+')
    def t_newline(self, t):
        t.lexer.lineno += len(t.value)
        return



from collections import OrderedDict

from .schema import RelationField


class DjangoQLSchemaSerializer(object):
    def serialize(self, schema):
        models = {}
        for model_label, fields in schema.models.items():
            models[model_label] = OrderedDict(
                [(name, self.serialize_field(f)) for name, f in fields.items()],
            )
        return {
            'current_model': schema.model_label(schema.current_model),
            'models': models,
        }

    def serialize_field(self, field):
        result = {
            'type': field.type,
            'nullable': field.nullable,
            'options': self.serialize_field_options(field),
        }
        if isinstance(field, RelationField):
            result['relation'] = field.relation
        return result

    def serialize_field_options(self, field):
        return list(field.get_options('')) if field.suggest_options else None


class SuggestionsAPISerializer(DjangoQLSchemaSerializer):
    def __init__(self, suggestions_api_url):
        self.suggestions_api_url = suggestions_api_url

    def serialize(self, schema):
        result = super(SuggestionsAPISerializer, self).serialize(schema)
        result['suggestions_api_url'] = self.suggestions_api_url
        return result

    def serialize_field_options(self, field):
        if field.async_options:
            return field.suggest_options
        else:
            return field.get_options('')



import time

from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.models import Group, User
from django.db.models import Count, Q
from django.utils.timezone import now

from djangoql.admin import DjangoQLSearchMixin
from djangoql.schema import DjangoQLSchema, IntField, StrField

from .models import Book


admin.site.unregister(User)


class ZaibatsuAdminSite(admin.AdminSite):
    site_header = 'Zaibatsu Admin'
    site_title = 'Zaibatsu Admin Portal'
    index_title = 'Welcome to Zaibatsu Admin Portal'


zaibatsu_admin_site = ZaibatsuAdminSite(name='zaibatsu')


class AuthorField(StrField):
    name = 'author'
    model = Book
    suggest_options = True

    def get_lookup_name(self):
        return 'author__username'

    def get_options(self, search):
        return Book.objects\
            .filter(author__username__icontains=search)\
            .values_list('author__username', flat=True)\
            .order_by('author__username')\
            .distinct()


class BookQLSchema(DjangoQLSchema):
    suggest_options = {
        Book: ['genre'],
    }

    def get_fields(self, model):
        fields = super(BookQLSchema, self).get_fields(model)
        if model == Book:
            fields += [AuthorField()]
        return fields


@admin.register(Book)
class BookAdmin(DjangoQLSearchMixin, admin.ModelAdmin):
    djangoql_schema = BookQLSchema
    list_display = ('name', 'author', 'genre', 'written', 'is_published')
    list_filter = ('is_published',)
    filter_horizontal = ('similar_books',)


class UserAgeField(IntField):
    """
    Search by given number of full years
    """
    model = User
    name = 'age'

    def get_lookup_name(self):
        """
        We'll be doing comparisons vs. this model field
        """
        return 'date_joined'

    def get_lookup(self, path, operator, value):
        if operator == 'in':
            result = None
            for year in value:
                condition = self.get_lookup(path, '=', year)
                result = condition if result is None else result | condition
            return result
        elif operator == 'not in':
            result = None
            for year in value:
                condition = self.get_lookup(path, '!=', year)
                result = condition if result is None else result & condition
            return result

        value = self.get_lookup_value(value)
        search_field = '__'.join(path + [self.get_lookup_name()])
        year_start = self.years_ago(value + 1)
        year_end = self.years_ago(value)
        if operator == '=':
            return (
                Q(**{'%s__gt' % search_field: year_start}) &
                Q(**{'%s__lte' % search_field: year_end})
            )
        elif operator == '!=':
            return (
                Q(**{'%s__lte' % search_field: year_start}) |
                Q(**{'%s__gt' % search_field: year_end})
            )
        elif operator == '>':
            return Q(**{'%s__lt' % search_field: year_start})
        elif operator == '>=':
            return Q(**{'%s__lt' % search_field: year_end})
        elif operator == '<':
            return Q(**{'%s__gt' % search_field: year_end})
        elif operator == '<=':
            return Q(**{'%s__gte' % search_field: year_start})

    def years_ago(self, n):
        timestamp = now()
        try:
            return timestamp.replace(year=timestamp.year - n)
        except ValueError:
            # February 29
            return timestamp.replace(month=2, day=28, year=timestamp.year - n)


class BookGenreField(StrField):
    model = Book
    name = 'name'
    suggest_options = True

    def get_options(self, search):
        time.sleep(1)
        return Book.objects.filter(
            name__icontains=search,
        ).values_list('name', flat=True)


class UserQLSchema(DjangoQLSchema):
    suggest_options = {
        Book: ['genre'],
        Group: ['name'],
    }

    def get_fields(self, model):
        fields = super(UserQLSchema, self).get_fields(model)
        if model == User:
            fields += [UserAgeField(), IntField(name='groups_count')]
        elif model == Book:
            fields += [BookGenreField()]
        return fields


class CustomUserAdmin(DjangoQLSearchMixin, UserAdmin):
    djangoql_schema = UserQLSchema
    search_fields = ('username', 'first_name', 'last_name')

    list_display = ('username', 'first_name', 'last_name', 'is_staff', 'group')

    def group(self, obj):
        return ', '.join([g.name for g in obj.groups.all()])
    group.short_description = 'Groups'

    def get_queryset(self, request):
        qs = super(CustomUserAdmin, self).get_queryset(request)
        return qs.\
            annotate(groups_count=Count('groups')).\
            prefetch_related('groups')


admin.site.register(User, CustomUserAdmin)
zaibatsu_admin_site.register(User, CustomUserAdmin)



from unittest import TestCase

from djangoql.ast import Comparison, Const, Expression, Name


class DjangoQLASTTest(TestCase):
    def test_equality(self):
        self.assertEqual(
            Expression(Name('age'), Comparison('='), Const(18)),
            Expression(Name('age'), Comparison('='), Const(18)),
        )
        self.assertNotEqual(
            Expression(Name('age'), Comparison('='), Const(42)),
            Expression(Name('age'), Comparison('='), Const(18)),
        )



from django.db.models import QuerySet

from .ast import Logical
from .parser import DjangoQLParser
from .schema import DjangoQLField, DjangoQLSchema


def build_filter(expr, schema_instance):
    if isinstance(expr.operator, Logical):
        left = build_filter(expr.left, schema_instance)
        right = build_filter(expr.right, schema_instance)
        if expr.operator.operator == 'or':
            return left | right
        else:
            return left & right

    field = schema_instance.resolve_name(expr.left)
    if not field:
        # That must be a reference to a model without specifying a field.
        # Let's construct an abstract lookup field for it
        field = DjangoQLField(
            name=expr.left.parts[-1],
            nullable=True,
        )
    return field.get_lookup(
        path=expr.left.parts[:-1],
        operator=expr.operator.operator,
        value=expr.right.value,
    )


def apply_search(queryset, search, schema=None):
    """
    Applies search written in DjangoQL mini-language to given queryset
    """
    ast = DjangoQLParser().parse(search)
    schema = schema or DjangoQLSchema
    schema_instance = schema(queryset.model)
    schema_instance.validate(ast)
    return queryset.filter(build_filter(ast, schema_instance))


class DjangoQLQuerySet(QuerySet):
    djangoql_schema = None

    def djangoql(self, search, schema=None):
        return apply_search(self, search, schema=schema or self.djangoql_schema)



import json

from django.contrib.auth.models import User
from django.test import TestCase


try:
    from django.core.urlresolvers import reverse
except ImportError:  # Django 2.0
    from django.urls import reverse


class DjangoQLAdminTest(TestCase):
    def setUp(self):
        self.credentials = {'username': 'test', 'password': 'lol'}
        User.objects.create_superuser(email='herp@derp.rr', **self.credentials)

    def get_json(self, url, status=200, **kwargs):
        response = self.client.get(url, **kwargs)
        self.assertEqual(status, response.status_code)
        try:
            return json.loads(response.content.decode('utf8'))
        except ValueError:
            self.fail('Not a valid json')

    def test_introspections(self):
        url = reverse('admin:core_book_djangoql_introspect')
        # unauthorized request should be redirected
        response = self.client.get(url)
        self.assertEqual(302, response.status_code)
        self.assertTrue(self.client.login(**self.credentials))
        # authorized request should be served
        introspections = self.get_json(url)
        self.assertEqual('core.book', introspections['current_model'])
        for model in ('core.book', 'auth.user', 'auth.group'):
            self.assertIn(model, introspections['models'])

    def test_introspection_suggestion_api_url(self):
        self.assertTrue(self.client.login(**self.credentials))
        for app in ['admin', 'zaibatsu']:
            url = reverse('%s:auth_user_djangoql_introspect' % app)
            introspections = self.get_json(url)
            self.assertEqual(
                reverse('%s:auth_user_djangoql_suggestions' % app),
                introspections['suggestions_api_url'],
            )

    def test_djangoql_syntax_help(self):
        for app in ['admin', 'zaibatsu']:
            url = reverse('%s:djangoql_syntax_help' % app)
            # unauthorized request should be redirected
            response = self.client.get(url)
            self.assertEqual(302, response.status_code)
            self.assertTrue(self.client.login(**self.credentials))
            # authorized request should be served
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)
            self.client.logout()

    def test_suggestions(self):
        url = reverse('admin:core_book_djangoql_suggestions')
        # unauthorized request should be redirected
        response = self.client.get(url)
        self.assertEqual(302, response.status_code)
        # authorize for the next checks
        self.assertTrue(self.client.login(**self.credentials))

        # field parameter is mandatory
        r = self.get_json(url, status=400)
        self.assertEqual(r.get('error'), '"field" parameter is required')

        # check for unknown fields
        r = self.get_json(url, status=400, data={'field': 'gav'})
        self.assertEqual(r.get('error'), 'Unknown field: gav')
        r = self.get_json(url, status=400, data={'field': 'x.y'})
        self.assertEqual(r.get('error'), 'Unknown model: core.x')
        r = self.get_json(url, status=400, data={'field': 'auth.user.lol'})
        self.assertEqual(r.get('error'), 'Unknown field: lol')

        # field with choices
        r = self.get_json(url, data={'field': 'genre'})
        self.assertEqual(r, {
            'page': 1,
            'has_next': False,
            'items': ['Drama', 'Comics', 'Other'],
        })

        # test that search is working
        r = self.get_json(url, data={'field': 'genre', 'search': 'o'})
        self.assertEqual(r, {
            'page': 1,
            'has_next': False,
            'items': ['Comics', 'Other'],
        })

        # ensure that page parameter is checked correctly
        r = self.get_json(url, status=400, data={'field': 'genre', 'page': 'x'})
        self.assertEqual(
            r.get('error'),
            "invalid literal for int() with base 10: 'x'",
        )
        r = self.get_json(url, status=400, data={'field': 'genre', 'page': '0'})
        self.assertEqual(
            r.get('error'),
            'page must be an integer starting from 1',
        )

        # check that paging after results end works correctly
        r = self.get_json(url, data={'field': 'genre', 'page': 2})
        self.assertEqual(r, {
            'page': 2,
            'has_next': False,
            'items': [],
        })

    def test_query(self):
        url = reverse('admin:core_book_changelist') + '?q=price=0'
        self.assertTrue(self.client.login(**self.credentials))
        response = self.client.get(url)
        # There should be no error at least
        self.assertEqual(200, response.status_code)



"""test_project URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.10/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
from django.conf import settings
from django.conf.urls import include
from django.contrib import admin


try:
    from django.urls import re_path  # Django >= 4.0
except ImportError:
    try:
        from django.conf.urls import re_path  # Django < 4.0
    except ImportError:  # Django < 2.0
        from django.conf.urls import url as re_path

from core.admin import zaibatsu_admin_site
from core.views import completion_demo


urlpatterns = [
    re_path(r'^admin/', admin.site.urls),
    re_path(r'^zaibatsu-admin/', zaibatsu_admin_site.urls),
    re_path(r'^$', completion_demo),
]

if settings.DEBUG and settings.DJDT:
    import debug_toolbar
    urlpatterns = [
        re_path(r'^__debug__/', include(debug_toolbar.urls)),
    ] + urlpatterns



import json

from django.contrib.auth.models import Group, User
from django.shortcuts import render
from django.views.decorators.http import require_GET

from djangoql.exceptions import DjangoQLError
from djangoql.queryset import apply_search
from djangoql.schema import DjangoQLSchema
from djangoql.serializers import DjangoQLSchemaSerializer


class UserQLSchema(DjangoQLSchema):
    include = (User, Group)
    suggest_options = {
        Group: ['name'],
    }


@require_GET
def completion_demo(request):
    q = request.GET.get('q', '')
    error = ''
    query = User.objects.all().order_by('username')
    if q:
        try:
            query = apply_search(query, q, schema=UserQLSchema)
        except DjangoQLError as e:
            query = query.none()
            error = str(e)
    # You may want to use SuggestionsAPISerializer and an additional API
    # endpoint (see in djangoql.views) for asynchronous suggestions loading
    introspections = DjangoQLSchemaSerializer().serialize(
        UserQLSchema(query.model),
    )
    return render(request, 'completion_demo.html', context={
        'q': q,
        'error': error,
        'search_results': query,
        'introspections': json.dumps(introspections),
    })



import json

from django.contrib import messages
from django.contrib.admin.views.main import ChangeList
from django.core.exceptions import FieldError, ValidationError
from django.db import DataError, NotSupportedError
from django.forms import Media
from django.http import HttpResponse
from django.template.loader import render_to_string
from django.views.generic import TemplateView

from .compat import text_type
from .exceptions import DjangoQLError
from .queryset import apply_search
from .schema import DjangoQLSchema
from .serializers import SuggestionsAPISerializer
from .views import SuggestionsAPIView


try:
    from django.core.urlresolvers import reverse
except ImportError:  # Django 2.0
    from django.urls import reverse

try:
    from django.urls import re_path  # Django >= 4.0
except ImportError:
    try:
        from django.conf.urls import re_path  # Django < 4.0
    except ImportError:  # Django < 2.0
        from django.conf.urls import url as re_path

DJANGOQL_SEARCH_MARKER = 'q-l'


class DjangoQLChangeList(ChangeList):
    def get_filters_params(self, *args, **kwargs):
        params = super(DjangoQLChangeList, self).get_filters_params(
            *args,
            **kwargs
        )
        if DJANGOQL_SEARCH_MARKER in params:
            del params[DJANGOQL_SEARCH_MARKER]
        return params


class DjangoQLSearchMixin(object):
    search_fields = ('_djangoql',)  # just a stub to have search input displayed
    djangoql_completion = True
    djangoql_completion_enabled_by_default = True
    djangoql_schema = DjangoQLSchema
    djangoql_syntax_help_template = 'djangoql/syntax_help.html'

    def search_mode_toggle_enabled(self):
        # If search fields were defined on a child ModelAdmin instance,
        # we suppose that the developer wants two search modes and therefore
        # enable search mode toggle
        return self.search_fields != DjangoQLSearchMixin.search_fields

    def djangoql_search_enabled(self, request):
        return request.GET.get(DJANGOQL_SEARCH_MARKER, '').lower() == 'on'

    def get_changelist(self, *args, **kwargs):
        return DjangoQLChangeList

    def get_search_results(self, request, queryset, search_term):
        if (
            self.search_mode_toggle_enabled() and
            not self.djangoql_search_enabled(request)
        ):
            return super(DjangoQLSearchMixin, self).get_search_results(
                request=request,
                queryset=queryset,
                search_term=search_term,
            )
        use_distinct = False
        if not search_term:
            return queryset, use_distinct

        try:
            qs = apply_search(queryset, search_term, self.djangoql_schema)
        except (DjangoQLError, ValueError, FieldError, ValidationError) as e:
            msg = self.djangoql_error_message(e)
            messages.add_message(request, messages.WARNING, msg)
            qs = queryset.none()
        else:
            # Hack to handle 'inet' comparison errors in Postgres. If you
            # know a better way to check for such an error, please submit a PR.
            try:
                # Django >= 2.1 has built-in .explain() method
                explain = getattr(qs, 'explain', None)
                if callable(explain):
                    try:
                        explain()
                    except NotSupportedError:
                        list(qs[:1])
                else:
                    list(qs[:1])
            except DataError as e:
                if 'inet' not in str(e):
                    raise
                msg = self.djangoql_error_message(e)
                messages.add_message(request, messages.WARNING, msg)
                qs = queryset.none()

        return qs, use_distinct

    def djangoql_error_message(self, exception):
        if isinstance(exception, ValidationError):
            msg = exception.messages[0]
        else:
            msg = text_type(exception)
        return render_to_string('djangoql/error_message.html', context={
            'error_message': msg,
            'djangoql_syntax_help_url': reverse(
                '%s:djangoql_syntax_help' % self.admin_site.name,
            ),
        })

    @property
    def media(self):
        media = super(DjangoQLSearchMixin, self).media
        if self.djangoql_completion:
            js = [
                'djangoql/js/completion.js',
            ]
            if self.search_mode_toggle_enabled():
                js.append('djangoql/js/completion_admin_toggle.js')
                if not self.djangoql_completion_enabled_by_default:
                    js.append('djangoql/js/completion_admin_toggle_off.js')
            js.append('djangoql/js/completion_admin.js')
            media += Media(
                css={'': (
                    'djangoql/css/completion.css',
                    'djangoql/css/completion_admin.css',
                )},
                js=js,
            )
        return media

    def get_urls(self):
        custom_urls = []
        if self.djangoql_completion:
            custom_urls += [
                re_path(
                    r'^introspect/$',
                    self.admin_site.admin_view(self.introspect),
                    name='%s_%s_djangoql_introspect' % (
                        self.model._meta.app_label,
                        self.model._meta.model_name,
                    ),
                ),
                re_path(
                    r'^suggestions/$',
                    self.admin_site.admin_view(self.suggestions),
                    name='%s_%s_djangoql_suggestions' % (
                        self.model._meta.app_label,
                        self.model._meta.model_name,
                    ),
                ),
                re_path(
                    r'^djangoql-syntax/$',
                    self.admin_site.admin_view(TemplateView.as_view(
                        template_name=self.djangoql_syntax_help_template,
                    )),
                    name='djangoql_syntax_help',
                ),
            ]
        return custom_urls + super(DjangoQLSearchMixin, self).get_urls()

    def introspect(self, request):
        suggestions_url = reverse('%s:%s_%s_djangoql_suggestions' % (
            self.admin_site.name,
            self.model._meta.app_label,
            self.model._meta.model_name,
        ))
        serializer = SuggestionsAPISerializer(suggestions_url)
        response = serializer.serialize(self.djangoql_schema(self.model))
        return HttpResponse(
            content=json.dumps(response, indent=2),
            content_type='application/json; charset=utf-8',
        )

    def suggestions(self, request):
        view = SuggestionsAPIView.as_view(
            schema=self.djangoql_schema(self.model),
        )
        return view(request)



# Generated by Django 2.1.2 on 2018-11-11 19:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0002_book_genre'),
    ]

    operations = [
        migrations.AddField(
            model_name='book',
            name='similar_books',
            field=models.ManyToManyField(to='core.Book', blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-31 20:46
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Book',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
                ('written', models.DateTimeField(default=django.utils.timezone.now)),
                ('is_published', models.BooleanField(default=False)),
                ('rating', models.FloatField(null=True)),
                ('price', models.DecimalField(decimal_places=2, max_digits=7, null=True)),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
    ]



class DjangoQLError(Exception):
    def __init__(self, message=None, value=None, line=None, column=None):
        self.value = value
        self.line = line
        self.column = column
        super(DjangoQLError, self).__init__(message)

    def __str__(self):
        message = super(DjangoQLError, self).__str__()
        if self.line:
            position_info = 'Line %s' % self.line
            if self.column:
                position_info += ', col %s' % self.column
            return '%s: %s' % (position_info, message)
        else:
            return message


class DjangoQLSyntaxError(DjangoQLError):
    pass


class DjangoQLLexerError(DjangoQLSyntaxError):
    pass


class DjangoQLParserError(DjangoQLSyntaxError):
    pass


class DjangoQLSchemaError(DjangoQLError):
    pass



# -*- coding: utf-8 -*-
import unittest.util
from unittest import TestCase

from djangoql.ast import Comparison, Const, Expression, List, Logical, Name
from djangoql.exceptions import DjangoQLParserError
from djangoql.parser import DjangoQLParser


# Show full contents in assertions when comparing long text strings
unittest.util._MAX_LENGTH = 2000


class DjangoQLParseTest(TestCase):
    parser = DjangoQLParser()

    def test_comparisons(self):
        self.assertEqual(
            Expression(Name('age'), Comparison('>='), Const(18)),
            self.parser.parse('age >= 18'),
        )
        self.assertEqual(
            Expression(Name('gender'), Comparison('='), Const('female')),
            self.parser.parse('gender = "female"'),
        )
        self.assertEqual(
            Expression(Name('name'), Comparison('!='), Const('Gennady')),
            self.parser.parse('name != "Gennady"'),
        )
        self.assertEqual(
            Expression(Name('married'), Comparison('in'),
                       List([Const(True), Const(False)])),
            self.parser.parse('married in (True, False)'),
        )
        self.assertEqual(
            Expression(Name('smile'), Comparison('!='), Const(None)),
            self.parser.parse('(smile != None)'),
        )
        self.assertEqual(
            Expression(Name(['job', 'best', 'title']), Comparison('>'),
                       Const('none')),
            self.parser.parse('job.best.title > "none"'),
        )

    def test_string_comparisons(self):
        self.assertEqual(
            Expression(Name('name'), Comparison('~'), Const('gav')),
            self.parser.parse('name ~ "gav"'),
        )
        self.assertEqual(
            Expression(Name('name'), Comparison('!~'), Const('gav')),
            self.parser.parse('name !~ "gav"'),
        )
        self.assertEqual(
            Expression(Name('name'), Comparison('startswith'), Const('gav')),
            self.parser.parse('name startswith "gav"'),
        )
        self.assertEqual(
            Expression(Name('name'), Comparison('not startswith'), Const('rr')),
            self.parser.parse('name not startswith "rr"'),
        )
        self.assertEqual(
            Expression(Name('name'), Comparison('endswith'), Const('gav')),
            self.parser.parse('name endswith "gav"'),
        )
        self.assertEqual(
            Expression(Name('name'), Comparison('not endswith'), Const('gav')),
            self.parser.parse('name not endswith "gav"'),
        )

    def test_escaped_chars(self):
        self.assertEqual(
            Expression(Name('name'), Comparison('~'),
                       Const(u'Contains a "quoted" str, 年年有余')),
            self.parser.parse(u'name ~ "Contains a \\"quoted\\" str, 年年有余"'),
        )
        self.assertEqual(
            Expression(Name('options'), Comparison('='), Const(u'П и Щ')),
            self.parser.parse(u'options = "\\u041f \\u0438 \\u0429"'),
        )

    def test_numbers(self):
        self.assertEqual(
            Expression(Name('pk'), Comparison('>'), Const(5)),
            self.parser.parse('pk > 5'),
        )
        self.assertEqual(
            Expression(Name('rating'), Comparison('<='), Const(523)),
            self.parser.parse('rating <= 5.23e2'),
        )

    def test_logical(self):
        self.assertEqual(
            Expression(
                Expression(Name('age'), Comparison('>='), Const(18)),
                Logical('and'),
                Expression(Name('age'), Comparison('<='), Const(45)),
            ),
            self.parser.parse('age >= 18 and age <= 45'),
        )
        self.assertEqual(
            Expression(
                Expression(
                    Expression(Name('city'), Comparison('='), Const('Ivanovo')),
                    Logical('and'),
                    Expression(Name('age'), Comparison('<='), Const(35)),
                ),
                Logical('or'),
                Expression(
                    Expression(Name('city'), Comparison('='), Const('Paris')),
                    Logical('and'),
                    Expression(Name('age'), Comparison('<='), Const(45)),
                ),
            ),
            self.parser.parse('(city = "Ivanovo" and age <= 35) or '
                              '(city = "Paris" and age <= 45)'),
        )

    def test_invalid_comparison(self):
        invalid_comparisons = (
            'foo > None',
            'b <= True',
            'c in False',
            '1 = 1',
            'a > b',
            'lol ~ None',
            'gav endswith 1',
            'nor not startswith False',
        )
        for expr in invalid_comparisons:
            self.assertRaises(DjangoQLParserError, self.parser.parse, expr)

    def test_entity_props(self):
        self.assertEqual(
            Expression(Name(['user', 'group', 'id']), Comparison('='),
                       Const(5)),
            self.parser.parse('user.group.id = 5'),
        )



from django.apps import apps
from django.contrib.auth.models import Group, User
from django.test import TestCase

from djangoql.exceptions import DjangoQLSchemaError
from djangoql.parser import DjangoQLParser
from djangoql.schema import DjangoQLSchema, IntField
from djangoql.serializers import SuggestionsAPISerializer

from ..models import Book


serializer = SuggestionsAPISerializer('/suggestions/')


class ExcludeUserSchema(DjangoQLSchema):
    exclude = (User,)


class IncludeUserGroupSchema(DjangoQLSchema):
    include = (Group, User)


class IncludeExcludeSchema(DjangoQLSchema):
    include = (Group,)
    exclude = (Book,)


class BookCustomFieldsSchema(DjangoQLSchema):
    def get_fields(self, model):
        if model == Book:
            return ['name', 'is_published']
        return super(BookCustomFieldsSchema, self).get_fields(model)


class WrittenInYearField(IntField):
    model = Book
    name = 'written_in_year'

    def get_lookup_name(self):
        return 'written__year'


class BookCustomSearchSchema(DjangoQLSchema):
    def get_fields(self, model):
        if model == Book:
            return [
                WrittenInYearField(),
            ]


class DjangoQLSchemaTest(TestCase):
    def all_models(self):
        models = []
        for app_label in apps.app_configs:
            models.extend(apps.get_app_config(app_label).get_models())
        return models

    def test_default(self):
        schema_dict = serializer.serialize(DjangoQLSchema(Book))
        self.assertIsInstance(schema_dict, dict)
        self.assertEqual('core.book', schema_dict.get('current_model'))
        models = schema_dict.get('models')
        self.assertIsInstance(models, dict)
        all_model_labels = sorted([str(m._meta) for m in self.all_models()])
        session_model = all_model_labels.pop()
        self.assertEqual('sessions.session', session_model)
        self.assertListEqual(all_model_labels, sorted(models.keys()))

    def test_exclude(self):
        schema_dict = serializer.serialize(ExcludeUserSchema(Book))
        self.assertEqual('core.book', schema_dict['current_model'])
        self.assertListEqual(sorted(schema_dict['models'].keys()), [
            'admin.logentry',
            'auth.group',
            'auth.permission',
            'contenttypes.contenttype',
            'core.book',
        ])

    def test_include(self):
        schema_dict = serializer.serialize(IncludeUserGroupSchema(User))
        self.assertEqual('auth.user', schema_dict['current_model'])
        self.assertListEqual(sorted(schema_dict['models'].keys()), [
            'auth.group',
            'auth.user',
        ])

    def test_get_fields(self):
        default_schema = DjangoQLSchema(Book)
        default = serializer.serialize(default_schema)['models']['core.book']
        custom_schema = BookCustomFieldsSchema(Book)
        custom = serializer.serialize(custom_schema)['models']['core.book']
        self.assertListEqual(list(default.keys()), [
            'author',
            'content_type',
            'genre',
            'id',
            'is_published',
            'name',
            'object_id',
            'price',
            'rating',
            'similar_books',
            'written',
        ])
        self.assertListEqual(list(custom.keys()), ['name', 'is_published'])

    def test_circular_references(self):
        models = serializer.serialize(DjangoQLSchema(Book))['models']
        # If Book references Author then Author should also reference Book back
        book_author_field = models['core.book'].get('author')
        self.assertIsNotNone(book_author_field)
        self.assertEqual('relation', book_author_field['type'])
        self.assertEqual('auth.user', book_author_field['relation'])
        self.assertEqual('relation', models['auth.user']['book']['type'])

    def test_custom_search(self):
        models = serializer.serialize(BookCustomSearchSchema(Book))['models']
        self.assertListEqual(
            list(models['core.book'].keys()),
            ['written_in_year'],
        )

    def test_invalid_config(self):
        try:
            IncludeExcludeSchema(Group)
            self.fail('Invalid schema with include & exclude raises no error')
        except DjangoQLSchemaError:
            pass
        try:
            IncludeUserGroupSchema(Book)
            self.fail('Schema was initialized with a model excluded from it')
        except DjangoQLSchemaError:
            pass
        try:
            IncludeUserGroupSchema(User())
            self.fail('Schema was initialized with an instance of a model')
        except DjangoQLSchemaError:
            pass

    def test_validation_pass(self):
        samples = [
            'first_name = "Lolita"',
            'groups.id < 42',
            'groups = None',  # that's ok to compare a model to None
            'groups != None',
            'groups.name in ("Stoics") and is_staff = False',
            'date_joined > "1753-01-01"',
            'date_joined > "1753-01-01 01:24"',
            'date_joined > "1753-01-01 01:24:42"',
        ]
        for query in samples:
            ast = DjangoQLParser().parse(query)
            try:
                IncludeUserGroupSchema(User).validate(ast)
            except DjangoQLSchemaError as e:
                self.fail(e)

    def test_validation_fail(self):
        samples = [
            'gav = 1',                      # unknown field
            'groups.gav > 1',               # unknown related field
            'groups = "lol"',               # can't compare model to a value
            'groups.name != 1',             # bad value type
            'is_staff = True and gav < 2',  # complex expression with valid part
            'date_joined < "1753-30-01"',   # bad timestamps
            'date_joined < "1753-01-01 12"',
            'date_joined < "1753-01-01 12AM"',
        ]
        for query in samples:
            ast = DjangoQLParser().parse(query)
            try:
                IncludeUserGroupSchema(User).validate(ast)
                self.fail("This query should't pass validation: %s" % query)
            except DjangoQLSchemaError:
                pass



#!/usr/bin/env python

from setuptools import setup

import djangoql


packages = ['djangoql']
requires = ['ply>=3.8']

setup(
    name='djangoql',
    version=djangoql.__version__,
    description='DjangoQL: Advanced search language for Django',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    author='Denis Stebunov',
    author_email='support@ivelum.com',
    url='https://github.com/ivelum/djangoql/',
    packages=packages,
    include_package_data=True,
    install_requires=requires,
    license=open('LICENSE').readline().strip(),
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
    ],
)



from django.contrib.auth.models import User
from django.test import TestCase, override_settings

from djangoql.queryset import apply_search
from djangoql.schema import DjangoQLSchema, IntField

from ..models import Book


class WrittenInYearField(IntField):
    model = Book
    name = 'written_in_year'

    def get_lookup_name(self):
        return 'written__year'


class BookCustomSearchSchema(DjangoQLSchema):
    suggest_options = {
        Book: ['genre'],
    }

    def get_fields(self, model):
        if model == Book:
            return [
                'genre', WrittenInYearField(),
            ]


class DjangoQLQuerySetTest(TestCase):
    def do_simple_query_test(self):
        qs = Book.objects.djangoql(
            'name = "foo" and author.email = "em@il" '
            'and written > "2017-01-30"',
        )
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual(
            '("core_book"."name" = foo AND "auth_user"."email" = em@il '
            'AND "core_book"."written" > 2017-01-30 00:00:00)',
            where_clause,
        )

    def test_simple_query(self):
        self.do_simple_query_test()

    @override_settings(USE_TZ=False)
    def test_simple_query_without_tz(self):
        self.do_simple_query_test()

    def test_datetime_like_query(self):
        qs = Book.objects.djangoql('written ~ "2017-01-30"')
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual(
            '"core_book"."written" LIKE %2017-01-30% ESCAPE \'\\\'',
            where_clause,
        )

    def test_advanced_string_comparison(self):
        qs = Book.objects.djangoql('name ~ "war"')
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual(
            '"core_book"."name" LIKE %war% ESCAPE \'\\\'',
            where_clause,
        )
        qs = Book.objects.djangoql('name startswith "war"')
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual(
            '"core_book"."name" LIKE war% ESCAPE \'\\\'',
            where_clause,
        )
        qs = Book.objects.djangoql('name not endswith "peace"')
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual(
            'NOT ("core_book"."name" LIKE %peace ESCAPE \'\\\')',
            where_clause,
        )

    def test_apply_search(self):
        qs = User.objects.all()
        try:
            qs = apply_search(qs, 'groups = None')
            qs.count()
        except Exception as e:
            self.fail(e)

    def test_choices(self):
        qs = Book.objects.djangoql(
            'genre = "Drama"',
            schema=BookCustomSearchSchema,
        )
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual('"core_book"."genre" = 1', where_clause)
        qs = Book.objects.djangoql(
            'genre in ("Drama", "Comics")',
            schema=BookCustomSearchSchema,
        )
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual('"core_book"."genre" IN (1, 2)', where_clause)

    def test_custom_field_query(self):
        qs = Book.objects.djangoql(
            'written_in_year = 2017',
            schema=BookCustomSearchSchema,
        )
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertTrue(
            where_clause.startswith('"core_book"."written" BETWEEN 2017-01-01'),
        )

    def test_empty_datetime(self):
        qs = apply_search(User.objects.all(), 'last_login = None')
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual('"auth_user"."last_login" IS NULL', where_clause)

    def test_datetime_in_query(self):
        qs = apply_search(User.objects.all(), 'last_login in ("2017-01-01")')
        where_clause = str(qs.query).split('WHERE')[1].strip()
        self.assertEqual(
            '"auth_user"."last_login" IN (2017-01-01 00:00:00)',
            where_clause,
        )
